@@ -30,10 +30,12 @@ static int get_page_vaddr(struct page_read *pr, struct iovec *iov)
 	return 1;
 }
 
-static int read_page(struct page_read *pr, unsigned long vaddr, void *buf)
+static int read_page(struct page_read *pr, unsigned long vaddr, int nr, void *buf)
 {
 	int ret;
 
+	BUG_ON(nr != 1);
+
 	ret = read(img_raw_fd(pr->pmi), buf, PAGE_SIZE);
 	if (ret != PAGE_SIZE) {
 		pr_err("Can't read mapping page %d\n", ret);
@@ -82,14 +84,14 @@ static void put_pagemap(struct page_read *pr)
 	pagemap_entry__free_unpacked(pr->pe, NULL);
 }
 
-static int read_pagemap_page(struct page_read *pr, unsigned long vaddr, void *buf);
+static int read_pagemap_page(struct page_read *pr, unsigned long vaddr, int nr, void *buf);
 
 static void skip_pagemap_pages(struct page_read *pr, unsigned long len)
 {
 	if (!len)
 		return;
 
-	pr_debug("\tpr%u Skip %lx bytes from page-dump\n", pr->id, len);
+	pr_debug("\tpr%u Skip %lu bytes from page-dump\n", pr->id, len);
 	if (!pr->pe->in_parent)
 		lseek(img_raw_fd(pr->pi), len, SEEK_CUR);
 	pr->cvaddr += len;
@@ -110,7 +112,7 @@ int seek_pagemap_page(struct page_read *pr, unsigned long vaddr, bool warn)
 
 		if (vaddr < pr->cvaddr) {
 			if (warn)
-				pr_err("Missing %lu in parent pagemap, current iov: base=%lx,len=%zu\n",
+				pr_err("Missing %lx in parent pagemap, current iov: base=%lx,len=%zu\n",
 					vaddr, (unsigned long)iov.iov_base, iov.iov_len);
 			return 0;
 		}
@@ -132,38 +134,85 @@ new_pagemap:
 	}
 }
 
-static int read_pagemap_page(struct page_read *pr, unsigned long vaddr, void *buf)
+static inline void pagemap_bound_check(PagemapEntry *pe, unsigned long vaddr, int nr)
+{
+	if (vaddr < pe->vaddr || (vaddr - pe->vaddr) / PAGE_SIZE + nr > pe->nr_pages) {
+		pr_err("Page read err %"PRIx64":%u vs %lx:%u\n",
+				pe->vaddr, pe->nr_pages, vaddr, nr);
+		BUG();
+	}
+}
+
+static int read_pagemap_page(struct page_read *pr, unsigned long vaddr, int nr, void *buf)
 {
 	int ret;
+	unsigned long len = nr * PAGE_SIZE;
+
+	pr_info("pr%u Read %lx %u pages\n", pr->id, vaddr, nr);
+	pagemap_bound_check(pr->pe, vaddr, nr);
 
 	if (pr->pe->in_parent) {
-		pr_debug("\tpr%u Read page %lx from parent\n", pr->id, vaddr);
-		ret = seek_pagemap_page(pr->parent, vaddr, true);
-		if (ret <= 0)
-			return -1;
-		ret = read_pagemap_page(pr->parent, vaddr, buf);
-		if (ret == -1)
-			return ret;
+		struct page_read *ppr = pr->parent;
+
+		/*
+		 * Parent pagemap at this point entry may be shorter
+		 * than the current vaddr:nr needs, so we have to
+		 * carefully 'split' the vaddr:nr into pieces and go
+		 * to parent page-read with the longest requests it
+		 * can handle.
+		 */
+
+		do {
+			int p_nr;
+
+			pr_debug("\tpr%u Read from parent\n", pr->id);
+			ret = seek_pagemap_page(ppr, vaddr, true);
+			if (ret <= 0)
+				return -1;
+
+			/*
+			 * This is how many pages we have in the parent
+			 * page_read starting from vaddr. Go ahead and
+			 * read as much as we can.
+			 */
+			p_nr = ppr->pe->nr_pages - (vaddr - ppr->pe->vaddr) / PAGE_SIZE;
+			pr_info("\tparent has %u pages in\n", p_nr);
+			if (p_nr > nr)
+				p_nr = nr;
+
+			ret = read_pagemap_page(ppr, vaddr, p_nr, buf);
+			if (ret == -1)
+				return ret;
+
+			/*
+			 * OK, let's see how much data we have left and go
+			 * to parent page-read again for the next pagemap
+			 * entry.
+			 */
+			nr -= p_nr;
+			vaddr += p_nr * PAGE_SIZE;
+			buf += p_nr * PAGE_SIZE;
+		} while (nr);
 	} else {
 		int fd = img_raw_fd(pr->pi);
 		off_t current_vaddr = lseek(fd, 0, SEEK_CUR);
-		pr_debug("\tpr%u Read page %lx from self %lx/%"PRIx64"\n", pr->id,
-				vaddr, pr->cvaddr, current_vaddr);
-		ret = read(fd, buf, PAGE_SIZE);
-		if (ret != PAGE_SIZE) {
+
+		pr_debug("\tpr%u Read page from self %lx/%"PRIx64"\n", pr->id, pr->cvaddr, current_vaddr);
+		ret = read(fd, buf, len);
+		if (ret != len) {
 			pr_perror("Can't read mapping page %d", ret);
 			return -1;
 		}
 
 		if (opts.auto_dedup) {
-			ret = punch_hole(pr, current_vaddr, (unsigned int)PAGE_SIZE, false);
+			ret = punch_hole(pr, current_vaddr, len, false);
 			if (ret == -1) {
 				return -1;
 			}
 		}
 	}
 
-	pr->cvaddr += PAGE_SIZE;
+	pr->cvaddr += len;
 
 	return 1;
 }
@@ -277,7 +326,7 @@ int open_page_read_at(int dfd, int pid, struct page_read *pr, int pr_flags)
 
 	pr->get_pagemap = get_pagemap;
 	pr->put_pagemap = put_pagemap;
-	pr->read_page = read_pagemap_page;
+	pr->read_pages = read_pagemap_page;
 	pr->close = close_page_read;
 	pr->id = ids++;
 
@@ -298,7 +347,7 @@ open_old:
 
 	pr->get_pagemap = get_page_vaddr;
 	pr->put_pagemap = NULL;
-	pr->read_page = read_page;
+	pr->read_pages = read_page;
 	pr->pi = NULL;
 	pr->close = close_page_read;
 
